/*
  mat4x4.c: sample code for vpu deivces
 
 	Copyright (C) 2000  Sony Computer Entertainment Inc.
 
  This file is subject to the terms and conditions of the GNU Library
  General Public License Version 2. See the file "COPYING.LIB" in the 
  main directory of this archive for more details.
*/


#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include "vpu_file.h"
#include "vpu_dev.h"
#include "vpu_event.h"
#include "mat4x4.h"



void
mat4x4_set(struct mat4x4 *mc)
{
	int i,j,k;

	for (i=0;i<4;i++) {
	    for (j=0;j<4;j++)
	       mc->val[i][j] = random()/(random()+1);
	}
}

void
mat4x4_show(struct mat4x4 *mc)
{
	int i,j,k;
	for (i=0;i<4;i++) {
	    for (j=0;j<4;j++)
	       printf(" %8.8g", mc->val[i][j]);
	    printf("\n");
	}
}

void
mat4x4_mul(struct mat4x4 *mc, struct mat4x4 *ma, struct mat4x4 *mb)
{
	int i,j,k;

	memset(mc,sizeof(*mc), 0);
	for (k=0;k<4;k++)
	   for (j=0;j<4;j++)
		for (i=0;i<4;i++) 
			mc->val[i][j]=mc->val[i][j]
					+ma->val[i][k]*mb->val[k][j];
}


void
mat4x4_mul_cop2(struct mat4x4 *mc, struct mat4x4 *ma, struct mat4x4 *mb)
{


	// MC = MA x MB
	
	asm __volatile__(
		// vf4.xyzw = mb[0].xyzw
		// vf5.xyzw = mb[1].xyzw
		// vf6.xyzw = mb[2].xyzw
		// vf7.xyzw = mb[3].xyzw

		"lqc2    vf4, (%2);
		 lqc2    vf5, 0x10(%2);
		 lqc2    vf6, 0x20(%2);
		 lqc2    vf7, 0x30(%2);"

		// i = 0;
		"move    $8, $0;"

	// do {
	"1:"
		// vf8.xyzw = ma[i].xyzw
		"lqc2    vf8, (%1);"

		// ACC.xyzw =  ma[i].x * mb[0].xyzw;
		// ACC.xyzw += ma[i].y * mb[1].xyzw;
		// ACC.xyzw += ma[i].z * mb[2].xyzw;
		// mc[i].xyzw = ACC.xzyw + ma[i].w * mb[3].xyzw;

		"vmulax.xyzw     ACC, vf4, vf8x;
		 vmadday.xyzw    ACC, vf5, vf8y;
		 vmaddaz.xyzw    ACC, vf6, vf8z;
		 vmaddw.xyzw     vf9, vf7, vf8w;"

		"sqc2    vf9, (%0);"

		// i++;
		"add    $8, 1;"

		//
		"add    %1, 0x10;"
		"add    %0, 0x10;"

	// } while (i < 4 ) ;
		"bne    $8, 4, 1b;"

		: "+r"(mc), "+r"(ma)	/* out or in-out */
		: "r" (mb)  		/* in */
		: "$8"/*t0*/ 		/* clobber */ 
	);
}


void
mat4x4_sub_cop2(struct mat4x4 *mc, struct mat4x4 *ma, struct mat4x4 *mb)
{


	// MC = MA - MB
	
	asm __volatile__(


		// i = 0;
		"move    $8, $0;"

	// do {
	"1:"

		"lqc2    vf4, (%1);"
		"lqc2    vf5, (%2);"
		"vsub.xyzw	vf6, vf4, vf5;\n"
		"sqc2    vf6, (%0);"

		// i++;
		"add    $8, 1;"

		// inc pointers
		"add    %2, 0x10;"
		"add    %1, 0x10;"
		"add    %0, 0x10;"

	// } while (i < 4 ) ;
		"bne    $8, 4, 1b"

		: "+r" (mc), "+r" (ma), "+r" (mb)  /* out or in-out */
		: 				/* in */
		: "$8"	  /*t0*/ 		/* clobber */ 
	);
}

/* return zero, if given two matrix is same */
int
mat4x4_cmp(struct mat4x4 *ma, struct mat4x4 *mb)
{

	int i,j;
	double a, b;
	double delta;
	double ep = pow(2,-23);

	for (i=0;i<4;i++)
	   for (j=0;j<4;j++) {
		a = ma->val[i][j] ;
		b = mb->val[i][j] ;
		delta = fabs(a*a-b*b)/(a*a+b*b);
		if ((a!=0.0 || b!=0.0) && (delta > ep )) {
			printf("%g:%g:%g\n", a, b, delta);
			return 1;
		}
	    }
	return 0;
}



#define N 10

struct mat4x4 *
mat4x4_mult_vpu(
	VPUDEV *vdev, 
	VPUFILE *vfd,
	struct mat4x4 *ma,
	struct mat4x4 *mb,
	int do_wait)
{
	int i;
	unsigned long *mult_data;
	unsigned long *prog_start;
	unsigned long *real_mult_data;
	unsigned long *map_mult_data;
	unsigned long *real_prog_start;
	void *start;
	unsigned int size;
	void *dstart;
	unsigned int dsize;
	int sect;
	void *vumem;
	void *umem;

	unsigned long *param;
	unsigned long *map_param;

	// Get VPU device mapped address
	umem=vpudev_text(vdev);
	vumem=vpudev_data(vdev);

	// Get mapped object address
	vpuobj_text(vfd, &start, &size);
	vpuobj_data(vfd, &dstart, &dsize);

	// Copy VPU text to VPU program memory.
	memcpy(umem, start, size);

	// Copy VPU data to vumemory.
	memcpy(vumem, dstart, dsize);


#define OBJ_TO_UMEM(x)	(void *)((unsigned int)(x)-(unsigned int)start)
#define OBJ_TO_UMAP(x)	(void *)((unsigned int)OBJ_TO_UMEM((x))+(unsigned int)umem)
#define OBJ_TO_VUMEM(x) (void *)((unsigned int)(x)-(unsigned int)dstart)
#define OBJ_TO_VUMAP(x)	(void *)((unsigned int)OBJ_TO_VUMEM((x))+(unsigned int)vumem)

	// Get param and data addr.
	i = vpuobj_symbol_by_name(vfd, "vpu_mult_data", 
				(void *)&mult_data, &sect);
	if ( i<0  || sect != 1) {
		fprintf(stderr,"symbol not found(vpu_mult_data)\n");
		exit(1);
	}
	i = vpuobj_symbol_by_name(vfd, "vpu_mult_param", 
				(void *)&param, &sect);
	if ( i<0  || sect != 1) {
		fprintf(stderr,"symbol not found(vpu_mult_param)\n");
		exit(1);
	}

	real_mult_data = OBJ_TO_VUMEM(mult_data);

	map_param = OBJ_TO_VUMAP(param);
	map_mult_data = OBJ_TO_VUMAP(mult_data);

	// Copy data
	memcpy(map_mult_data, ma, sizeof(*ma)*N);
	memcpy((void *)map_mult_data+sizeof(*ma)*N, mb, sizeof(*mb)*N);

	// Set params.
	// 	x: pointer to results array 
	map_param[0]= ((unsigned int)real_mult_data + 
				sizeof(*ma)*N + sizeof(*mb)*N);
	map_param[0] /= 16;
	// 	y: pointer to array A
	map_param[1]= ((unsigned int)real_mult_data);
	map_param[1] /= 16;
	// 	z: pointer to array B
	map_param[2]= ((unsigned int)real_mult_data + sizeof(*mb)*N);
	map_param[2] /= 16;
	// 	w: number of elements
	map_param[3]= N;

	// Get VPU prog addr (.vutext)
	i = vpuobj_symbol_by_name(vfd, "vpu_mult", (void *)&prog_start, &sect);
	if ( i<0  || sect != 0) {
		fprintf(stderr,"symbol not found(vpu_mutl)\n");
		exit(1);
	}
	real_prog_start  = OBJ_TO_UMEM(prog_start); 
	
	// Start VPU prog.
	vpudev_start(vdev, (unsigned int)real_prog_start);

	// wait VPU prog. stop
	if (do_wait)
		while(vpudev_busy(vdev)) {
			usleep(1);
	}

	// return 

	return  (struct mat4x4 *)
		((unsigned int)map_mult_data + sizeof(*ma)*N + sizeof(*mb)*N);

}


int
mat4x4_test(
	VPUDEV *vdev0, 
	VPUDEV *vdev1, 
	VPUFILE *vfd)
{

	int evfd;
	int i;
	int total_ng;
	int ng;
	struct timeval tv;

	static struct mat4x4 a [N] __attribute__((aligned(16)));
	static struct mat4x4 b [N] __attribute__((aligned(16)));
	static struct mat4x4 c [N] __attribute__((aligned(16)));
	static struct mat4x4 d [N] __attribute__((aligned(16)));
	static struct mat4x4 *p;
	static struct mat4x4 *q;


	// Init randome seed
	gettimeofday(&tv, 0);
	srandom(tv.tv_usec);


	// Init matrix data
	for (i=0;i<N;i++) {
		mat4x4_set(&a[i]);
		mat4x4_set(&b[i]);

	}

	// open event device
	evfd=vpuevent_open();

	// Do mult at VPU0 macro mode
	printf("* Do 4x4 matrix mult at VPU0 (macro mode)\n");
	p = mat4x4_mult_vpu(vdev0, vfd,  &a[0], &b[0], 0);

	// Do mult at VPU1 macro mode
	printf("* Do 4x4 matrix mult at VPU1 (macro mode)\n");
	q = mat4x4_mult_vpu(vdev1, vfd,  &a[0], &b[0], 0);

	// Do mult at EE core
	printf("* Do 4x4 matrix mult at EE core\n");
	q = mat4x4_mult_vpu(vdev1, vfd,  &a[0], &b[0], 0);
	for (i=0;i<N;i++) {
		mat4x4_mul(&c[i], &a[i], &b[i]);
	}

	// Wait VPU0 and VPU1
	vpuevent_wait(evfd);
	// close event device
	vpuevent_close(evfd);

	// Do mult at COP2 (VPU1 micro mode)
	printf("* Do 4x4 matrix mult at COP2 (VPU0 micro mode)\n");
	for (i=0;i<N;i++) {
		mat4x4_mul_cop2(&d[i], &a[i], &b[i]);
	}

	// Check result
	total_ng = 0;
	ng = 0;
	printf("* Check VPU0 micro mode result\n");
	for (i=0;i<N;i++) {
		int j;
		j = mat4x4_cmp(&d[i], &c[i]);
		if (j) {
			mat4x4_show(&d[i]);
			mat4x4_show(&c[i]);
			printf("\n");
			ng ++;
		}
	}
	if (ng==0) printf("Passed\n");

	total_ng += ng;
	ng = 0;
	printf("* Check VPU0 macro mode result\n");
	for (i=0;i<N;i++) {
		int j;
		j = mat4x4_cmp(&p[i], &c[i]);
		if (j) {
			mat4x4_show(&p[i]);
			mat4x4_show(&c[i]);
			printf("\n");
			ng ++;
		}
	}
	if (ng==0) printf("Passed\n");

	total_ng += ng;
	ng = 0;
	printf("* Check VPU1 macro mode result\n");
	for (i=0;i<N;i++) {
		int j;
		j = mat4x4_cmp(&q[i], &c[i]);
		if (j) {
			mat4x4_show(&q[i]);
			mat4x4_show(&c[i]);
			printf("\n");
			ng ++;
		}
	}
	if (ng==0) printf("Passed\n");
	total_ng += ng;

	return total_ng;
}

