/*
  vpu_dev.c: sample code for vpu deivces
 
 	Copyright (C) 2000  Sony Computer Entertainment Inc.
 
  This file is subject to the terms and conditions of the GNU Library
  General Public License Version 2. See the file "COPYING.LIB" in the 
  main directory of this archive for more details.
*/

#include <sys/types.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <linux/ps2/dev.h>
#include <errno.h>
#include "vpu_dev.h"


struct vpudev_struct {
	int fd;
	int vpu;
	unsigned int mapsize;
	void *text;
	unsigned int text_size;
	void *data;
	unsigned int data_size;
};


VPUDEV vpudev[2] = {{fd:-1,}, {fd:-1}};

void vpudev_reset(VPUDEV *dev)
{

	__asm__ __volatile__(
		"cfc2.	$8," CP2_FBRST ";"	// read from FBRST
		"ctc2.	%0," CP2_FBRST ";"	// write to FBRST
		"sync.p;"
		"or	$8, %1;"
		"ctc2.	$8," CP2_FBRST ";"	// write to FBRST
		"sync.p;"
	: 
	: "r" (dev->vpu?RS1:RS0), "r" (dev->vpu?(TE1|DE1):(TE0|DE0))
	: "$8"
	);
}

void vpudev_start(VPUDEV *dev, unsigned int start)
{
	start &= 0xffff;
	
	if (dev->vpu) 
		__asm__ __volatile__(
			"ctc2	%0," CP2_CMSAR1 ";"	// write to CMSAR1
			"sync.p;"
			"vnop;"
		:
		: "r" (start)
		);
	else
		__asm__ __volatile__(
			"ctc2	%0," CP2_CMSAR0 ";"	// write to CMSAR0
			"sync.p;"
			"vnop;"
			"vcallmsr vi27"
		:
		: "r" (start)
		);

}

int vpudev_busy(VPUDEV *dev)
{
	register int reg=0;

	__asm__ __volatile__(
		"cfc2	%0," CP2_VPU_STAT";"	// read from VPU-STAT
		"and	%0, %1;"
	: "=r" (reg)
	: "r" (dev->vpu?VBS1:VBS0)
	);
	return reg;
}


VPUDEV *
vpudev_open(int vpu)
{
	int i;
	int fd;
	static char err[100];
	struct ps2_vpuinfo vpuinfo;
	void *vputext_addr;
	void *vpudata_addr;
	char *dev_name;
	void *ptr;
	int mapsize;
	VPUDEV * pvpu;

	pvpu = vpu ? &vpudev[1] : &vpudev[0];
	dev_name = vpu ? PS2_DEV_VPU1: PS2_DEV_VPU0;

	if (pvpu->fd != -1) {
		errno = EBUSY;
		return (void *)-1;
	}

	fd = open(dev_name, O_RDWR);
	if (fd<0){
		sprintf(err, "%s:open", dev_name);
		perror(err);
		return (void *)-1;
	}

	i = ioctl(fd, PS2IOC_VPUINFO, &vpuinfo);
	if (i<0) {
		sprintf(err, "%s:PS2IOC_VPUINFO", dev_name);
		perror(err);
		close(fd);
		return (void *)-1;
	}
	mapsize = vpuinfo.umemsize+vpuinfo.vumemsize;

	ptr = mmap(NULL, mapsize, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	if (ptr == (void*) -1 ) {
		sprintf(err, "%s:mmap", dev_name);
		perror(err);
		close(fd);
		return (void *)-1;
	}

	pvpu->vpu = vpu;
	pvpu->fd = fd;
	pvpu->text = ptr;
	pvpu->text_size = vpuinfo.umemsize;
	pvpu->data = ptr+vpuinfo.umemsize;
	pvpu->data_size = vpuinfo.vumemsize;
	pvpu->mapsize = mapsize;

	return pvpu;

}

void * vpudev_text(VPUDEV *pvpu)
{
	return pvpu->text;
}

void * vpudev_data(VPUDEV *pvpu)
{
	return pvpu->data;
}

void
vpudev_close(VPUDEV *pvpu)
{
	munmap(pvpu->text, pvpu->mapsize);
	close(pvpu->fd);
	pvpu->fd = -1;
}

